def func_best_model(scan_object,
                    metric='val_acc',
                    asc=False,
                    saved=False,
                    custom_objects=None):

    '''Picks the best model based on a given metric and
    returns the index number for the model.

    scan_object | object | The object that is returned from Scan()
                           upon completion.
    metric | str | The metric to be used for picking the best model.
    asc | bool | Make this True for metrics that are to be minimized
                 (e.g. loss), and False when the metric is to be
                 maximized (e.g. acc).
    saved | bool | if a model saved on local machine should be used.
    custom_objects | dict | if the model has a custom object, pass it here.

    NOTE: for loss 'asc' should be True'''

    import warnings as warnings

    warnings.simplefilter('ignore')

    from ..utils.best_model import best_model, activate_model
    model_no = best_model(scan_object, metric, asc)
    out = activate_model(scan_object, model_no, saved, custom_objects)

    return out


def func_evaluate(scan_object,
                  x_val,
                  y_val,
                  task,
                  n_models=10,
                  metric='val_acc',
                  folds=5,
                  shuffle=True,
                  asc=False,
                  saved=False,
                  custom_objects=None):

    '''K-fold Cross Evaluator

    For creating scores from kfold cross-evaluation and
    adding them to the data frame.

    scan_object | python class | The class object returned by Scan() upon
                                 completion of the experiment.
    x_val | array or list of arrays | Input data (features) in the same format
                                      as used in Scan(), but should not be the
                                      same data (or it will not be much of
                                      validation).
    y_val | array or list of arrays | Input data (labels) in the same format
                                      as used in.
                                      Scan(), but should not be the same data
                                      (or it will not be much of validation).
    task | string | 'binary', 'multi_class', 'multi_label', or 'continuous'.
    n_models | int | The number of models to be evaluated. If set to 10,
                     then 10 models with the highest metric value are
                     evaluated. See below.
    metric | str | The metric to be used for picking the models to be
                   evaluated.
    folds | int | The number of folds to be used in the evaluation.
    shuffle | bool | If the data is to be shuffled or not. Set always
                     to False for timeseries but keep in mind that you
                     might get periodical/seasonal bias.
    asc | bool | Set to True if the metric is to be minimized.
    saved | bool | if a model saved on local machine should be used
    custom_objects | dict | if the model has a custom object, pass it here

    '''
    import warnings as warnings
    from tqdm import tqdm
    import numpy as np

    warnings.simplefilter('ignore')

    picks = scan_object.data.sort_values(metric,
                                         ascending=asc).index.values[:n_models]

    if n_models > len(scan_object.data):
        data_len = len(scan_object.data)
    else:
        data_len = n_models

    out = []

    pbar = tqdm(total=data_len)

    from ..commands.evaluate import Evaluate

    for i in range(len(scan_object.data)):

        if i in list(picks):

            evaluate_object = Evaluate(scan_object)
            temp = evaluate_object.evaluate(x_val,
                                            y_val,
                                            task=task,
                                            model_id=i,
                                            metric=metric,
                                            folds=folds,
                                            shuffle=shuffle,
                                            asc=asc,
                                            saved=saved,
                                            custom_objects=custom_objects)

            out.append([np.mean(temp), np.std(temp)])
            pbar.update(1)

        else:
            out.append([np.nan, np.nan])

    pbar.close()

    if task == 'continuous':
        heading = 'eval_' + 'mae'
    else:
        heading = 'eval_' + 'f1score'

    scan_object.data[heading + '_mean'] = [i[0] for i in out]
    scan_object.data[heading + '_std'] = [i[1] for i in out]

    print(">> Added evaluation score columns to scan_object.data")



def power_draw_append(history, power_draw):

    '''For appending the data from PowerDrawCallback to the history object
    and allowing the data to be captured in the experiment log in Talos.

    history | object | tf.keras model history object
    power_draw | object | PowerDrawCallback object

    '''

    import numpy as np

    joined = power_draw.log['epoch_begin'] + power_draw.log['epoch_end']
    history.history['watts_min'] = [min(joined)]
    history.history['watts_max'] = [max(joined)]
    history.history['seconds'] = [sum(power_draw.log['seconds'])]

    # get average watts per epoc
    epoch_begin = np.array(power_draw.log['epoch_begin'])
    epoch_end = np.array(power_draw.log['epoch_end'])
    avg_watts = (epoch_begin + epoch_end) / 2

    watt_seconds = round(sum(avg_watts * np.array(power_draw.log['seconds'])), 2)
    history.history['Ws'] = [watt_seconds]

    return history



class Predict:

    '''Class for making predictions on the models that are stored
    in the Scan() object'''

    def __init__(self, scan_object):

        '''Takes in as input a Scan() object and returns and object
        with properties for `predict` and `predict_classes`'''

        self.scan_object = scan_object
        self.data = scan_object.data

    def predict(self,
                x,
                metric,
                asc,
                model_id=None,
                saved=False,
                custom_objects=None):

        '''Makes a probability prediction from input x. If model_id
        is not given, then best_model will be used.

        x | array | data to be used for the predictions
        model_id | int | the id of the model from the Scan() object
        metric | str | the metric to be used for picking best model
        asc | bool | True if `metric` is something to be minimized
        saved | bool | if a model saved on local machine should be used
        custom_objects | dict | if the model has a custom object,
                                pass it here

        '''

        if model_id is None:
            from ..utils.best_model import best_model
            model_id = best_model(self.scan_object, metric, asc)

        from ..utils.best_model import activate_model
        model = activate_model(self.scan_object,
                               model_id,
                               saved,
                               custom_objects)

        return model.predict(x)

    def predict_classes(self,
                        x,
                        metric,
                        asc,
                        model_id=None,
                        saved=False,
                        custom_objects=None):

        '''Makes a class prediction from input x. If model_id
        is not given, then best_model will be used.

        x | array | data to be used for the predictions
        model_id | int | the id of the model from the Scan() object
        metric | str | the metric to be used for picking best model
        asc | bool | True if `metric` is something to be minimized
        saved | bool | if a model saved on local machine should be used
        custom_objects | dict | if the model has a custom object, pass it here
        '''

        import numpy as np

        if model_id is None:
            from ..utils.best_model import best_model
            model_id = best_model(self.scan_object, metric, asc)

        from ..utils.best_model import activate_model
        model = activate_model(self.scan_object,
                               model_id,
                               saved,
                               custom_objects)

        # make (class) predictions with the model
        preds = model.predict(x)   
        preds_classes = np.argmax(preds, axis=1)

        return preds_classes



class DistributeParamSpace:

    def __init__(self,
                 params,
                 param_keys,
                 random_method='uniform_mersenne',
                 fraction_limit=None,
                 round_limit=None,
                 time_limit=None,
                 boolean_limit=None,
                 machines=2):

        '''Splits ParamSpace object based on number
        of machines.

        params | object | ParamSpace class object
        machines | int | number of machines to split for

        NOTE: `Scan()` limits will not be applied if ParamSpace object
        is passed directly into `Scan()` as `params` argument so they
        should be passed directly into `DistributeParamSpace` instead.

        '''

        from talos.parameters.ParamSpace import ParamSpace

        self._params = ParamSpace(params=params,
                                  param_keys=param_keys,
                                  random_method='uniform_mersenne',
                                  fraction_limit=None,
                                  round_limit=None,
                                  time_limit=None,
                                  boolean_limit=None)

        self.machines = machines

        self.param_spaces = self._split_param_space()

    def _split_param_space(self):

        '''Takes in a ParamSpace object and splits it so that
        it can be used in DistributeScan experiments.'''

        import numpy as np
        import copy

        out = {}

        # randomly shuffle the param_space
        rand = np.random.default_rng()
        rand.shuffle(self._params.param_space, axis=0)

        # split into n arras
        param_spaces = np.array_split(self._params.param_space, self.machines)

        # remove keys to allow copy
        param_keys = self._params.param_keys
        self._params.param_keys = []

        # create the individual ParamSpace objects
        for i in range(self.machines):

            out[i] = copy.deepcopy(self._params)
            out[i].param_space = param_spaces[i]
            out[i].dimensions = len(out[i].param_space)
            out[i].param_index = list(range(out[i].dimensions))
            out[i].param_keys = param_keys

        return out



def best_model(self, metric, asc):

    '''Picks the best model based on a given metric and
    returns the index number for the model.

    NOTE: for loss 'asc' should be True'''

    best = self.data.sort_values(metric, ascending=asc).iloc[0].name

    return best


def activate_model(self, model_id, saved=False, custom_objects=None):

    '''Loads the model from the json that is stored in the Scan object
    or from local

    model_id | int | the sequential id of the model
    saved | bool | if a model saved on local machine should be used
    custom_object | dict | if the model has a custom object, pass it here

    '''

    import tensorflow as tf
    from tensorflow.keras.models import model_from_json

    if saved:

        file_path = self.details['experiment_name']
        file_path += '/' + self.details['experiment_id']
        file_path += '/' + str(model_id)

        model = tf.keras.models.load_model(file_path,
                                           custom_objects=custom_objects)

    else:
        model = model_from_json(self.saved_models[model_id])
        model.set_weights(self.saved_weights[model_id])

    return model
